      subroutine LU_Decomposition(A_input_matrix,b_input_vector,length,
     &                            A_LU_equivalent_matrix,i_MLS_step,
     &                            i_Singular,x_solution,i_in)

      !Solves Ax = b using LU Decomposition
      !A = LU  ==> LUx = b
      !Ux = y  ==> Ly  = b
      !Solve Ly = b using forward subsitution
      !Solve Ux = y using backward substitution
      
      parameter(LU_size_max = 4)
      real A_input_matrix(LU_size_max,LU_size_max)
      real A_temp(LU_size_max,LU_size_max)
      real A_LU_equivalent_matrix(LU_size_max,LU_size_max)
      real L_matrix(LU_size_max,LU_size_max)
      real U_matrix(LU_size_max,LU_size_max)
      real x_solution(LU_size_max), y_solution(LU_size_max)
      real b_input_vector(LU_size_max), sum_Uijxj, sum_Lijyj
      TOL_REAL = 1.0e-8
      TOL_DBLE = 1.0e-16

      if(i_MLS_step.eq.1)then

       !- Check singularity -
       amax_Aij = 0.0
       row_sum_min = 10.0e8
       i_Singular = 0
       do i = 1,length
         row_sum = 0.0
         do j=1,length
           !Transfer input matrix into a temporary matrix
           A_temp(i,j) = A_input_matrix(i,j)
           !Check matrix is not singular
           A_temp_ij_abs = abs(A_temp(i,j))
           row_sum = row_sum + A_temp_ij_abs
           if(A_temp_ij_abs.gt.amax_Aij)then
             amax_Aij = A_temp_ij_abs
           endif
           if(row_sum.lt.row_sum_min)then
             row_sum_min = row_sum
           endif
         enddo
       enddo
       if(amax_Aij   .lt.TOL_REAL) i_Singular = 1
       if(row_sum_min.lt.TOL_DBLE) i_Singular = 1
       
       
       if(i_Singular.lt.1)then
         do i = 1,length
           L_matrix(i,1)=A_input_matrix(i,1)
         enddo
         do j = 1,length
           U_matrix(1,j)=A_input_matrix(1,j) / L_matrix(1,1)
         enddo
         !- Forward elimination using Gaussian Elimination for matrices L and U -
         do j = 2,length
           do i = j,length
             SUM = 0.0
             do k = 1,j-1
               SUM = SUM + L_matrix(i,k)*U_matrix(k,j)
             enddo
             L_matrix(i,j) = A_input_matrix(i,j) - SUM
           enddo
           U_matrix(j,j) = 1.0
           do i = j+1,length
             sum = 0.0
             do k = 1,j-1
               sum = sum + L_matrix(j,k)*U_matrix(k,i)
             enddo
             U_matrix(j,i) = (A_input_matrix(j,i) - SUM)/L_matrix(j,j)
           enddo
         enddo 
 
         
         !- Combine into one matrix -
         do i = 1,length  !Row
           do j = i+1,length !Column
               A_LU_equivalent_matrix(i,j) = U_matrix(i,j)
           enddo
           do j = 1,i !Column
               A_LU_equivalent_matrix(i,j) = L_matrix(i,j)
           enddo
         enddo
       
       endif
  
      else  !if(i_MLS_step.eq.2)then  !Solver equation

      x_solution = 0.0  !zero initialise solution vector
      y_solution = 0.0  !zero initialise temp     vector
      !- Forward Substitution using Lower Matrix -
      y_solution(1)=b_input_vector(1)/A_LU_equivalent_matrix(1,1) !First equation has only one unknown
      do i = 2,length
        sum_Lijyj = 0.0
        !- Generating the summation term -
        do j = 1,i-1
          sum_Lijyj = sum_Lijyj
     &              + A_LU_equivalent_matrix(i,j)*y_solution(j)
        enddo
        y_solution(i)=(b_input_vector(i)- sum_Lijyj) /
     &                      A_LU_equivalent_matrix(i,i)
      enddo
  
  
      !- Back Substitution -
      x_solution(length) = y_solution(length)!Nth equation has only one unknown
      do i = length-1,1,-1
        sum_Uijxj = 0.0
        !- Generating the summation term -
        do j = i+1,length
          sum_Uijxj = sum_Uijxj
     &              + A_LU_equivalent_matrix(i,j)*x_solution(j)
        enddo
        x_solution(i) = (y_solution(i) - sum_Uijxj)
      enddo
       
      endif   !End of:        if(i_MLS_step.eq.1)then

      end subroutine
